import io
from contextlib import redirect_stdout, redirect_stderr
from pydantic import BaseModel
from multiprocessing import Process, Pipe
from bondai.tools import Tool

DEFAULT_EXECUTION_TIMEOUT = 60
TOOL_NAME = 'python_repl'
TOOL_DESCRIPTION = (
    "This tool allows you to execute Python code. "
    "Specify your Python code in the 'code' parameter and it will return the result."
)

class Parameters(BaseModel):
    code: str
    thought: str

def execute_target(conn, code):
    local_vars = {}
    stdout_str, stderr_str = "", ""
    
    try:
        with io.StringIO() as stdout_io, io.StringIO() as stderr_io, redirect_stdout(stdout_io), redirect_stderr(stderr_io):
            exec(code, {}, local_vars)
            stdout_str = stdout_io.getvalue()
            stderr_str = stderr_io.getvalue()

        # Remove non-picklable objects from local_vars if any
        for key in list(local_vars.keys()):
            if not isinstance(local_vars[key], (int, float, str, list, dict, tuple)):
                del local_vars[key]
                
        conn.send([local_vars, stdout_str, stderr_str])
    except Exception as e:
        conn.send([str(e), stdout_str, stderr_str])

class PythonREPLTool(Tool):
    def __init__(self, execution_timeout=DEFAULT_EXECUTION_TIMEOUT):
        super(PythonREPLTool, self).__init__(TOOL_NAME, TOOL_DESCRIPTION, parameters=Parameters, dangerous=True)
        self.execution_timeout = execution_timeout
    
    def run(self, arguments):
        code = arguments.get('code')

        if code is None:
            raise Exception("'code' parameter is required")

        result, stdout, stderr = self.execute_code(code)
        
        response = ""
        
        # Include stdout if present
        if stdout:
            response += f"Output:\n{stdout}\n"
            
        # Include stderr if present
        if stderr:
            response += f"Errors:\n{stderr}\n"
        
        # Include result if present
        if result:
            formatted_result = "\n".join([f"{key}: {value}" for key, value in result.items()])
            response += f"Result Variables:\n{formatted_result}\n"

        if not response:
            response = "Code executed successfully. No output or result variables."
        
        return response
    
    def execute_code(self, code):
        # Create a pipe for communication
        parent_conn, child_conn = Pipe()

        process = Process(target=execute_target, args=(child_conn, code))
        process.start()
        process.join(timeout=self.execution_timeout)

        if process.is_alive():
            process.terminate()
            process.join(10)
            raise Exception("Code execution timed out")

        result, stdout, stderr = parent_conn.recv()
        if isinstance(result, str):
            raise Exception(result)

        return result, stdout, stderr











from pydantic import BaseModel
from bondai.tools import Tool
from bondai.models.openai import OpenAILLM, MODEL_GPT4_0613

TOOL_NAME = 'agent_tool'
TOOL_DESCRIPTION = (
    "This tool allows you to delegate tasks to other agents. "
    "This can be really helpful for taking a complex task and breaking it down into smaller, more manageable pieces. "
    "Just include a highly descriptive prompt in the 'task_description' parameter for this task. "
    "The more detailed your description the better the agent will be at the task. "
    "The 'task_description' parameter is required and MUST be provided."
)

class Parameters(BaseModel):
    task_description: str
    thought: str

class AgentTool(Tool):
    def __init__(self, agent=None, prompt_builder=None, tools=[], llm=OpenAILLM(MODEL_GPT4_0613)):
        super(AgentTool, self).__init__(TOOL_NAME, TOOL_DESCRIPTION, Parameters)
        
        from bondai import Agent
        if agent:
            self.agent = agent
        else:
            self.agent = Agent(prompt_builder=prompt_builder, tools=tools, llm=llm)
    
    def run(self, arguments):
        task_description = arguments.get('task_description')
        if task_description is None:
            raise Exception("'task_description' is required.")
        
        result = self.agent.run(task_description).output
        self.agent.reset_memory()
        return result



from bondai.api import BondAIAPIClient

task = "I want you to write a story about unicorns and save it to a file named unicorns.md."

# Create the client
client = BondAIAPIClient()

# Connect to the WebSocket
client.connect_ws()

# Listen to WebSocket events
@client.on("agent_message")
def handle_agent_message(message):
    print(message)
    print("Enter your response:")
    response=input()
    client.send_user_message(response)

@client.on("agent_started")
def handle_agent_started():
    print('Agent has started.')

@client.on("agent_step_completed")
def handle_agent_step_completed(message):
    print("Agent step completed.")
    print(message)

@client.on("agent_completed")
def handle_agent_completed():
    print('Agent has completed.')

# Start the Agent
response = client.start_agent()

try:
    while True:
        pass
except KeyboardInterrupt:
    print("Exiting...")